"""
Module for defining protocols for type checking.

The protocols defined here match the properties of the classes in `onboardapis.mixins`.
"""
from __future__ import annotations

from typing import Protocol

from .data import ID, StationType, Position
from .mixins import InternetAccessInterface


class SupportsPosition(Protocol):
    # noinspection PyPropertyDefinition
    @property
    def position(self) -> Position: ...


class SupportsSpeed(Protocol):
    # noinspection PyPropertyDefinition
    @property
    def speed(self) -> float: ...


class SupportsStations(Protocol):
    def calculate_distance(self, station: StationType): ...

    # noinspection PyPropertyDefinition
    @property
    def stations_dict(self) -> dict[ID, StationType]: ...

    # noinspection PyPropertyDefinition
    @property
    def stations(self) -> list[StationType]: ...

    # noinspection PyPropertyDefinition
    @property
    def origin(self) -> StationType: ...

    # noinspection PyPropertyDefinition
    @property
    def current_station(self) -> StationType: ...

    # noinspection PyPropertyDefinition
    @property
    def destination(self) -> StationType: ...

    # noinspection PyPropertyDefinition
    @property
    def delay(self) -> float: ...

    # noinspection PyPropertyDefinition
    @property
    def distance(self) -> float: ...


class SupportsInternetAccess(Protocol):
    # noinspection PyPropertyDefinition
    @property
    def internet_access(self) -> InternetAccessInterface: ...



"""
Package for trains

## Operator ID

The operator ID for trains is different depending on the region.

For Europe, the operator ID is the [VKM register code](https://www.era.europa.eu/domains/registers/vkm_en).

---

Contains the following countries:

- ``onboardapis.train.at`` - Austria
- ``onboardapis.train.de`` - Germany
- ``onboardapis.train.fr`` - France
- ``onboardapis.train.it`` - Italy

---
"""
from __future__ import annotations

import logging

from abc import ABCMeta
from dataclasses import dataclass
from typing import cast

from .. import Vehicle, Station, ConnectingVehicle
from ..data import ScheduledEvent

# noinspection PyUnresolvedReferences
from typing import Iterable  # for associations while generating docs
# noinspection PyUnresolvedReferences
from ..data import ID, Position  # for associations while generating docs
# noinspection PyUnresolvedReferences
from datetime import datetime  # for associations while generating docs

logger = logging.getLogger(__name__)


__all__ = [
    "at",
    "de",
    "fr",
    "it",
    "third_party",
    "TrainStation",
    "Train",
    "ConnectingTrain",
]


@dataclass
class TrainStation(Station):
    """
    An `onboardapis.Station` with the additional information of a platform
    """
    _connections: Iterable[ConnectingTrain]

    platform: ScheduledEvent[str] | None
    """The platform where the train arrives."""

    @property
    def connections(self) -> list[ConnectingTrain]:
        return cast(list[ConnectingTrain], super().connections)  # maybe force type instead of casting?


class Train(Vehicle, metaclass=ABCMeta):
    """
    Base class for all train implementations.
    """

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    @property
    def type(self) -> str:
        """The abbreviated train type."""
        return 'undefined'

    @property
    def line_number(self) -> str:
        """The line number of this train."""
        return 'undefined'


@dataclass
class ConnectingTrain(ConnectingVehicle):
    """
    A connecting train is a train that is not part of the main trip but of a connecting service

    It may only have limited information available
    """

    platform: ScheduledEvent[str] | None
    """
    The platform where the train will depart from
    """



from __future__ import annotations

import json
import logging

from typing import ClassVar
from urllib.parse import parse_qs, urlparse

import requests

from gql import gql
from requests.exceptions import ConnectionError

from ....exceptions import InitialConnectionError
from ....data import ThreadedGraphQlAPI, get_package_version
from ....mixins import InternetAccessInterface

logger = logging.getLogger(__name__)


class GenericUnwiredAPI(ThreadedGraphQlAPI):
    API_URL = 'https://wasabi-splashpage.wifi.unwired.at/api/graphql'
    UNWIRED_GRAPHQL_QUERY: ClassVar[str] = '''
    query feed_widget($user_session_id: ID, $ap_mac: String, $widget_id: ID!, $language: String) {
          feed_widget(
            user_session_id: $user_session_id
            ap_mac: $ap_mac
            widget_id: $widget_id
            language: $language
          ) {
            user_session_id
            error {
              ...Error
              __typename
            }
            widget {
              ...Widget
              __typename
            }
            __typename
          }
        }
        fragment Error on Error {
          error_code
          error_message
          __typename
        }
        fragment Widget on Widget {
          widget_id
          page_id
          position
          date_updated
          ... on SimpleTextWidget {
            is_ready
            ...SimpleTextWidget
            __typename
          }
          ... on ConnectWidget {
            button_text
            connected_text
            variant
            confirmation
            delay
            require_sms_auth
            email_mandatory
            terms_of_service
            store_terms
            enable_anchor
            anchor {
              ...Anchor
              __typename
            }
            __typename
          }
          ... on JourneyInfoWidget {
            json
            enable_anchor
            anchor {
              ...Anchor
              __typename
            }
            variant
            is_ready
            hold_text
            __typename
          }
          ... on StructuredTextWidget {
            is_ready
            categories {
              ...StructuredTextCategory
              __typename
            }
            __typename
          }
          ... on SupportFormWidget {
            custom_options {
              option_key
              text
              email
              __typename
            }
            __typename
          }
          ... on Wifi4EUWidget {
            self_test
            network_identifier
            __typename
          }
          ... on EmergencyRequestWidget {
            reasons {
              reason
              __typename
            }
            disclaimer
            status
            __typename
          }
          ... on MovingMapWidget {
            is_ready
            icon
            geo_points {
              icon_width
              icon_url
              lat
              long
              text
              __typename
            }
            json
            __typename
          }
          __typename
        }
        fragment Anchor on Anchor {
          slug
          label
          __typename
        }
        fragment SimpleTextWidget on SimpleTextWidget {
          content
          enable_anchor
          anchor {
            ...Anchor
            __typename
          }
          __typename
        }
        fragment StructuredTextCategory on StructuredTextCategory {
          label
          entries {
            ...StructuredTextEntry
            __typename
          }
          enable_anchor
          anchor {
            ...Anchor
            __typename
          }
          __typename
        }
        fragment StructuredTextEntry on StructuredTextEntry {
          title
          content
          POI_match {
            ...PoiMatch
            __typename
          }
          __typename
        }
        fragment PoiMatch on PoiMatch {
          stop {
            name
            id
            ds100
            ibnr
            __typename
          }
          __typename
        }
    '''
    """The query to execute against the GraphQL API."""

    _user_session_id: str

    def __init__(self):
        ThreadedGraphQlAPI.__init__(self)

    def init(self):
        try:
            response = requests.get(
                'https://unwired.info/?source=wasabi',
                headers={'User-Agent': 'Python/onboardapis (%s)' % get_package_version()}
            )
            self._user_session_id, *_ = parse_qs(urlparse(response.url).query)['user_session_id']
        except (ConnectionError, KeyError) as e:
            raise InitialConnectionError from e

    def refresh(self) -> None:
        response = self.execute(
            document=gql(self.UNWIRED_GRAPHQL_QUERY),
            variable_values={
                "widget_id": "363a8707-5e3e-4a5b-b565-9d22470dfd25",
                "language": "de",
                "user_session_id": self._user_session_id,
            }
        )

        journey = json.loads(response['feed_widget']['widget']['json'])
        self._data['journey'] = (
            {} if 'content' in journey.keys()  # No journey info found for your location!
            else journey['course']
        )


class GenericUnwiredInternetAccessInterface(InternetAccessInterface):
    def enable(self) -> None:  # TODO: implement
        pass

    def disable(self) -> None:  # TODO: implement
        pass



"""
Package for other vehicle types

Other vehicle types could be anything that doesn't fit into the other categories.
They may be additions that will be added in the future or vehicles that are not
common enough to warrant their own category.

This package is also a namespace package that contains no code
to allow for extension packages.
"""
from __future__ import annotations
from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)  # noqa



"""
Implementation for Icomera AB.

Contains the following APIs:

- ``GenericIcomeraTrain``

---
"""
from __future__ import annotations

from .apis import GenericIcomeraTrain
from .interfaces import GenericIcomeraAPI, GenericIcomeraInternetAccessInterface

__all__ = [
    "GenericIcomeraTrain",
    "GenericIcomeraAPI",
    "GenericIcomeraInternetAccessInterface",
]



from __future__ import annotations

import logging

from ....units import meters_per_second
from ....data import Position
from ... import Train
from .interfaces import RailnetRegioAPI

logger = logging.getLogger(__name__)


class RailnetRegio(Train):
    """
    Wrapper for interacting with the ÖBB RailnetRegio API
    """

    _api: RailnetRegioAPI

    def __init__(self):
        self._api = RailnetRegioAPI()
        Train.__init__(self)

    @property
    def speed(self) -> float:
        return meters_per_second(kilometers_per_hour=self._api['gps']['JSON']['speed'])

    @property
    def position(self) -> Position:
        return Position(
            latitude=self._api['gps']['JSON']['lat'],
            longitude=self._api['gps']['JSON']['lon'],
            # these are only present in newer trains
            altitude=self._api['gps']['JSON'].get('alt', None),
            heading=self._api['gps']['JSON'].get('bearing', None),
        )



"""
Implementation for the German operator Deutsche Bahn / DB AG.

Contains the following APIs:

- ``ICEPortal``
- ``RegioGuide``

---
"""
from __future__ import annotations

from .apis import ICEPortal, RegioGuide, ZugPortal

__all__ = [
    "ICEPortal",
    "RegioGuide",
    "ZugPortal",
]



"""
Implementation for the German operator metronom Eisenbahngesellschaft mbH.

Contains the following APIs:

- ``MetronomCaptivePortal``

---
"""
from __future__ import annotations

from .apis import MetronomCaptivePortal

__all__ = [
    "MetronomCaptivePortal",
]



from __future__ import annotations

import logging

from abc import ABCMeta

from bs4 import BeautifulSoup

from ....data import BlockingRestAPI
from ....exceptions import APIConnectionError
from ....mixins import InternetAccessInterface

logger = logging.getLogger(__name__)


class GenericIcomeraAPI(BlockingRestAPI, metaclass=ABCMeta):
    # noinspection HttpUrlsUsage
    API_URL = 'http://omboard.info'


class GenericIcomeraInternetAccessInterface(InternetAccessInterface):
    _api: BlockingRestAPI

    def __init__(self, api: BlockingRestAPI) -> None:
        InternetAccessInterface.__init__(self, api)

    def enable(self) -> None:
        response = self._api.get('de')
        if BeautifulSoup(response.text, 'html.parser').find(class_='user-offline') is None:
            return  # Already online

        response = self._api.post('de/', data={
            'login': True,
            'CSRFToken': response.cookies['csrf'],
        })
        response.raise_for_status()
        if BeautifulSoup(response.text, 'html.parser').find(class_='user-online') is None:
            raise APIConnectionError('Login failed!')

    def disable(self) -> None:
        response = self._api.get('de')
        if BeautifulSoup(response.text, 'html.parser').find(class_='user-online') is None:
            return  # Already offline

        response = self._api.post('de/', data={
            'logout': True,
            'CSRFToken': response.cookies['csrf'],
        })
        response.raise_for_status()
        if BeautifulSoup(response.text, 'html.parser').find(class_='user-offline') is None:
            raise APIConnectionError('Logout failed!')

    @property
    def is_enabled(self) -> bool:
        return BeautifulSoup(
            self._api.get('de').text, 'html.parser'
        ).find(class_='user-online') is not None



from __future__ import annotations

import logging
import time

from functools import lru_cache
from typing import TypedDict

from deprecation import deprecated

from ....data import ThreadedRestAPI, store, get_package_version

logger = logging.getLogger(__name__)

Endpoints = TypedDict("Endpoints", {
    'api/train_info': int,
    'api/gps': dict,
    'api/speed': float,
})
"""A dict containing information about the endpoints of the Railnet API"""


class RailnetRegioAPI(ThreadedRestAPI):
    API_URL = "https://railnet.oebb.at"

    @store('train_info')
    @lru_cache
    def train_info(self) -> int:
        return int(self.get("api/trainInfo").text)

    @store('gps')
    def gps(self) -> dict:
        return self.get("api/gps").json()

    @store('speed')
    def speed(self) -> float:
        return float(self.get("api/speed").text)

    @deprecated(
        deprecated_in='2.0.0',
        removed_in='2.1.0',
        current_version=get_package_version(),
        details='combined has been removed from the API by ÖBB'
    )
    @store('combined')
    def combined(self) -> dict:
        return self.get("assets/modules/fis/combined.json", params={"_time": time.time()}).json()

    def refresh(self) -> None:
        self.train_info()
        self.gps()
        self.speed()



"""
Mixins for vehicles.

The mixins are used to indicate and add functionality to the vehicle classes.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from datetime import timedelta
from typing import Generic

from .exceptions import DataInvalidError
from .data import Position, ID, StationType, API


class PositionMixin(metaclass=ABCMeta):
    @property
    @abstractmethod
    def position(self) -> Position:
        """
        Get the current position of the vehicle.

        :return: The current position of the vehicle
        :rtype: Position
        """
        raise NotImplementedError


class SpeedMixin(metaclass=ABCMeta):
    """
    Functionality for a vehicle that provides information about its speed.
    """

    @property
    @abstractmethod
    def speed(self) -> float:
        r"""The current speed of the vehicle.

        Returns:
            The current speed in $\frac{meters}{second}$
        """
        raise NotImplementedError


class StationsMixin(Generic[StationType], metaclass=ABCMeta):
    """
    Functionality for a vehicle that provides information on the journey.
    """

    def calculate_distance(self, station: StationType) -> float:
        """Calculate the distance in meters between the train and a station.

        Use the trains ``position`` or ``distance`` to calculate the distance to ``station``.

        Args:
            station: The station to calculate the distance to

        Returns:
            The distance between the train and the station in meters
        """
        if hasattr(self, 'position'):
            position: Position = getattr(self, 'position')
            return station.calculate_distance(position)

        if hasattr(self, 'distance'):
            distance: float = getattr(self, 'distance')
            return station.calculate_distance(distance)

        raise NotImplementedError

    @property
    @abstractmethod
    def stations_dict(self) -> dict[ID, StationType]:
        """A dict containing the stations on the journey.

        The stations that this vehicle passes through returned as a dict with the station ID as the key.

        Returns:
            The stations as a dict
        """
        raise NotImplementedError

    @property
    def stations(self) -> list[StationType]:
        """Return the stations as a list.

        Return a list that contains every station from `onboardapis.mixins.StationsMixin.stations_dict`.

        Returns:
            The `stations_dict` values
        """
        return list(self.stations_dict.values())

    @property
    def origin(self) -> StationType:
        """The station where this vehicle started the current journey.

        Returns:
            The first station on this trip.
        """
        if len(self.stations) > 0:
            return self.stations[0]
        raise DataInvalidError("No origin station found")

    @property
    @abstractmethod
    def current_station(self) -> StationType:
        """The next station.

        Returns:
            The station where this vehicle will arrive next or is currently at.
        """
        raise NotImplementedError

    @property
    def destination(self) -> StationType:
        """The station where this vehicle terminates the current journey.

        Returns:
            The destination station.
        """
        if len(self.stations) > 0:
            return self.stations[-1]
        raise DataInvalidError("No destination station found")

    @property
    def delay(self) -> timedelta:
        """The current delay.

        The current delay calculated by the `current_station`
        scheduled arrival time vs. actual arrival time.

        Returns:
            The current delay of the vehicle as a `datetime.timedelta` object.
        """
        return timedelta(seconds=(
            self.current_station.arrival.actual - self.current_station.arrival.scheduled
        ).total_seconds())

    @property
    def is_delayed(self) -> bool:
        """Whether the train is delayed.

        Convenience function that tests if `delay > timedelta(0)`.
        """
        return self.delay > timedelta()

    @property
    def distance(self) -> float:
        """
        The distance from the start in meters

        :return: The distance
        :rtype: float
        """
        return self.calculate_distance(self.origin)


class InternetAccessMixin(metaclass=ABCMeta):
    """Adds the internet_access property to a class
    that defines an :class:`InternetAccessInterface` as ``_internet_access``."""
    _internet_access: InternetAccessInterface

    @property
    def internet_access(self) -> InternetAccessInterface:
        """An interface to manage the internet access for this device."""
        return self._internet_access


class InternetAccessInterface(metaclass=ABCMeta):
    """Interface adding functions for connecting and disconnecting to the internet
    as well as viewing the current status."""

    _is_enabled: bool = False
    """Cached information on connection status"""
    _api: API

    def __init__(self, api: API) -> None:
        self._api = api

    @abstractmethod
    def enable(self) -> None:
        """Enable the internet access for this device.

        Request internet access for this device by automatically accepting the terms of service
        and signing in to the captive portal.

        Raises:
            ConnectionError: If the internet access is temporarily not available.
        """
        self._is_enabled = True

    @abstractmethod
    def disable(self) -> None:
        """Disable the internet access for this device.

        Disable the internet access for this device by signing out of the captive portal.

        Raises:
            ConnectionError: If the internet access is temporarily not available.
        """
        if not self.is_enabled:
            return

        self._is_enabled = False

    @property
    def is_enabled(self) -> bool:
        """Return whether the internet access is enabled for this device."""
        return self._is_enabled


class InternetMetricsInterface(metaclass=ABCMeta):
    """Interface for information on limited internet access."""

    @abstractmethod
    def limit(self) -> int | None:
        """Return the total internet access quota in MB or `None` if there is none."""




"""
Package for planes

## Operator ID

The operator ID for planes is the
[3 letter ICAO airline designator](https://en.wikipedia.org/wiki/List_of_airline_codes#Codes).
"""
from __future__ import annotations

from abc import ABCMeta

from .. import Vehicle

__all__ = [
    "Plane",
]


class Plane(Vehicle, metaclass=ABCMeta):
    """
    Base class for planes
    """



"""
Convert between different units
"""

from __future__ import annotations

from geopy.units import kilometers, meters, miles, feet, nautical, km, m, mi, ft, nm

__all__ = [
    'kilometers',
    'meters',
    'miles',
    'feet',
    'nautical',
    'km',
    'm',
    'mi',
    'ft',
    'nm',
    'kmh',
    'ms',
    'seconds',
    'minutes',
    'hours',
    'coordinates_decimal_to_dms',
    'coordinates_dms_to_decimal',
    'kilometers_per_hour',
    'kilometers_per_hour',
    'meters_per_second',
    'meters_per_second',
]


def coordinates_decimal_to_dms(
        coordinates: tuple[float, float]
) -> tuple[tuple[int, int, float], tuple[int, int, float]]:
    """
    Convert the tuple ``coordinates`` of coordinates to degrees, minutes, seconds

    :param coordinates: The decimal coordinates to convert to degrees, minutes, seconds
    :type coordinates: Tuple[float, float]
    :return: The coordinates in degrees, minutes, seconds
    :rtype: Tuple[Tuple[int, int, float], Tuple[int, int, float]]
    """
    lat, lon = coordinates
    lat_deg = int(abs(lat))
    lat_min = int((abs(lat) - lat_deg) * 60)
    lat_sec = ((abs(lat) - lat_deg) * 60 - lat_min) * 60
    lon_deg = int(abs(lon))
    lon_min = int((abs(lon) - lon_deg) * 60)
    lon_sec = ((abs(lon) - lon_deg) * 60 - lon_min) * 60
    return (
        (-lat_deg if lat < 0 else lat_deg, lat_min, lat_sec),
        (-lon_deg if lon < 0 else lon_deg, lon_min, lon_sec)
    )


def coordinates_dms_to_decimal(
        coordinates: tuple[tuple[int, int, float], tuple[int, int, float]]
) -> tuple[float, float]:
    """
    Convert the tuple ``coordinates`` of degrees, minutes, seconds to decimal degrees

    :param coordinates: The degrees, minutes, seconds coordinates to convert to decimal degrees
    :type coordinates: Tuple[Tuple[int, int, float], Tuple[int, int, float]]
    :return: The coordinates in decimal degrees
    :rtype: Tuple[float, float]
    """
    (lat_deg, lat_min, lat_sec), (lon_deg, lon_min, lon_sec) = coordinates
    lat = lat_deg + lat_min / 60 + lat_sec / 3600
    lon = lon_deg + lon_min / 60 + lon_sec / 3600
    return lat, lon


def seconds(hours: float = 0, minutes: float = 0) -> float:  # noqa: F402
    """Convert to seconds"""
    ret = 0
    if hours:
        ret += hours * 3600
    if minutes:
        ret += minutes * 60
    return ret


def minutes(hours: float = 0, seconds: float = 0) -> float:  # noqa: F402
    """Convert to minutes"""
    ret = 0
    if hours:
        ret += hours * 60
    if seconds:
        ret += seconds / 60
    return ret


def hours(minutes: float = 0, seconds: float = 0) -> float:  # noqa: F402
    """Convert to hours"""
    ret = 0
    if minutes:
        ret += minutes / 60
    if seconds:
        ret += seconds / 3600
    return ret


def kilometers_per_hour(meters_per_second: float = None) -> float:  # noqa: F402
    """Convert to kilometers per hour"""
    ret = 0
    if meters_per_second:
        ret += kilometers(meters=meters_per_second) / hours(seconds=1)
    return ret


def meters_per_second(kilometers_per_hour: float = None) -> float:  # noqa: F402
    """Convert to meters per second"""
    ret = 0
    if kilometers_per_hour:
        ret += meters(kilometers=kilometers_per_hour) / seconds(hours=1)
    return ret


kmh = kilometers_per_hour
ms = meters_per_second



"""
Implementation of the Italian operator Trenitalia S.p.A.

Contains the following APIs:

- ``PortaleRegionale``

---
"""
from __future__ import annotations

from .apis import PortaleRegionale

__all__ = [
    "PortaleRegionale",
]



from __future__ import annotations

import logging

from datetime import datetime, timedelta
from typing import Dict, Any

from ....data import (
    Position,
    ScheduledEvent,
)
from ... import Train, TrainStation
from .interfaces import InouiConnector

logger = logging.getLogger(__name__)


class PortalINOUI(Train):
    _api: InouiConnector

    def __init__(self):
        self._api = InouiConnector()
        Train.__init__(self)

    @property
    def id(self) -> str:
        return self._api["details"].get("trainId")

    @property
    def type(self) -> str:
        return self._api["details"].get(
            "carrier"
        )  # todo carrier != type?

    @property
    def line_number(self) -> str:
        return self._api["details"].get("number")

    @property
    def stations_dict(self) -> Dict[Any, TrainStation]:
        return {
            stop.get("code"): TrainStation(
                id=stop.get("code"),
                name=stop.get("label"),
                platform=None,
                arrival=ScheduledEvent(
                    scheduled=datetime.fromisoformat(stop.get("theoricDate")),
                    actual=datetime.fromisoformat(stop.get("realDate")),
                ),
                departure=ScheduledEvent(
                    scheduled=(
                        datetime.fromisoformat(stop.get("theoricDate"))
                        + timedelta(minutes=int(stop.get("duration", 0)))
                    ),
                    actual=(
                        datetime.fromisoformat(stop.get("realDate"))
                        + timedelta(minutes=int(stop.get("duration", 0)))
                    ),
                ),
                position=Position(
                    latitude=int(stop.get("coordinates", {}).get("latitude", 0)),
                    longitude=int(stop.get("coordinates", {}).get("longitude", 0)),
                ),
                distance=float(stop.get("progress", {}).get("remainingDistance", 0)),
                _connections=self._api.connections(station_id=stop.get("code")),
            )
            for stop in self._api["details"].get("stops", [])
        }

    @property
    def current_station(self) -> TrainStation:
        current_station_id, *_ = (
            stop.get("code")
            for stop
            in self._api["details"].get("stops", [])
            if 0 < int(stop.get("progress", {}).get('progressPercentage', 0)) < 100
        )
        return self.stations_dict.get(current_station_id)

    @property
    def speed(self) -> float:
        return float(self._api["gps"].get("speed"))

    @property
    def distance(self) -> float:
        return sum(
            float(stop.get('progress', {}).get('traveledDistance', 0))
            for stop
            in self._api['details'].get('stops', [])
        )

    @property
    def position(self) -> Position:
        gps = self._api["gps"]
        return Position(
            latitude=float(gps.get("latitude", 0)),
            longitude=float(gps.get("longitude", 0)),
            altitude=float(gps.get("altitude", 0)),
            heading=float(gps.get("heading", 0)),
        )



from __future__ import annotations

import logging

from abc import ABCMeta

from ....mixins import InternetAccessMixin
from ... import Train
from .interfaces import GenericIcomeraAPI, GenericIcomeraInternetAccessInterface

logger = logging.getLogger(__name__)


class GenericIcomeraTrain(Train, InternetAccessMixin, metaclass=ABCMeta):
    _api: GenericIcomeraAPI
    _internet_access: GenericIcomeraInternetAccessInterface

    def __init__(self):
        if not hasattr(self, '_api'):
            self._api = GenericIcomeraAPI()
        if not hasattr(self, '_internet_access'):
            self._internet_access = GenericIcomeraInternetAccessInterface(self._api)
        Train.__init__(self)
        InternetAccessMixin.__init__(self)



"""
Implementation of the German operator Flix Train GmbH.

Contains the following APIs:

- ``FlixTainment``

---
"""
from __future__ import annotations

from .apis import FlixTainment

__all__ = [
    "FlixTainment"
]



"""
Module for data generation and processing.

This module contains classes and functions that are used to generate and process data fetched from the APIs.
It also provides common data structures that are used throughout the library.
"""
from __future__ import annotations

import importlib.metadata
import logging
import time

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from functools import wraps
from json import JSONDecodeError
from typing import TypeVar, Generic, ClassVar, Callable
from threading import Thread, Lock, Event

from geopy.point import Point
from geopy.distance import geodesic
from gql import Client
from gql.transport.requests import RequestsHTTPTransport
from restfly import APISession

from .units import coordinates_decimal_to_dms
from .exceptions import APIConnectionError, InitialConnectionError, APIFeatureMissingError

__all__ = [
    "ID",
    "StationType",
    "get_package_version",
    "default",
    "ScheduledEvent",
    "Position",
    "API",
    "store",
    "ThreadedAPI",
    "BlockingRestAPI",
    "ThreadedRestAPI",
    "BlockingGraphQlAPI",
    "ThreadedGraphQlAPI",
]

logger = logging.getLogger(__name__)

T = TypeVar("T")
"""A type variable for generics."""

ID = TypeVar("ID", str, int)
"""A TypeVar indicating the return type of Vehicle.id"""

StationType = TypeVar("StationType", bound="Station")
"""A TypeVar indicating the Station type"""


def get_package_version() -> str:
    """Return the version of the ``onboardapis`` package."""
    try:
        return importlib.metadata.version('onboardapis')
    except importlib.metadata.PackageNotFoundError:
        return 'unknown'


def default(arg: any, default: any = None, *, boolean: bool = True) -> any:  # noqa: F402
    """Return ``arg`` if it evaluates to ``True``, else return ``default``.

    Set ``boolean`` to ``False`` to only test for ``arg is None``.

    Args:
        arg: The data to test.
        default: The default value to return if no data is present.
        boolean: Return the default if `arg` evaluates to False.

    Returns:
        The data if present, else the default.
    """
    if boolean:
        return arg if arg else default
    return default if arg is None else arg


class ScheduledEvent(Generic[T]):
    """
    An event that is scheduled to have the value ``scheduled``,
    but can also happen different from the expected and actually happens as ``actual``.
    """

    scheduled: T
    """The expected value of this event."""
    actual: T
    """The actual value of this event."""

    def __init__(self, scheduled: T, actual: T | None = None) -> None:
        """
        Initialize a new :class:`ScheduledEvent`.

        Args:
            scheduled: The value that should happen.
            actual: The value that actually happens, will be the same as the scheduled value if passed as None.
        """
        self.scheduled = scheduled
        self.actual = actual or scheduled

    def __str__(self) -> str:
        return str(default(self.actual, self.scheduled))


@dataclass
class Position(object):
    """
    A position requires at least a latitude and longitude,
    but can also provide data on altitude and the current compass heading.
    """

    latitude: float
    """The latitude in decimal degrees"""
    longitude: float
    """The longitude in decimal degrees"""
    altitude: float = None
    """The altitude in meters"""
    heading: float = None
    """The compass heading in degrees"""

    def __str__(self) -> str:
        (lat_deg, lat_min, lat_sec), (lon_deg, lon_min, lon_sec,) = coordinates_decimal_to_dms(
            (self.latitude, self.longitude)
        )
        coordinates = (
            f"{abs(lat_deg)}°{lat_min}'{lat_sec:.3f}\"{'N' if lat_deg >= 0 else 'S'}"
            + f" {abs(lon_deg)}°{lon_min}'{lon_sec:.3f}\"{'E' if lon_deg >= 0 else 'W'}"
            + (f" {self.altitude:.2f}m" if self.altitude is not None else "")
            + (f" {self.heading:.2f}°" if self.heading is not None else "")
        )
        return coordinates

    def __getitem__(self, item) -> float:
        return (self.latitude, self.longitude)[item]

    def calculate_distance(self, other: Position | Point) -> float:
        """
        Calculate the distance (in meters) between this position and another position.

        :param other: The other position
        :return: The distance in meters
        """
        if not isinstance(other, (Position, Point)):
            raise ValueError
        return geodesic(self.to_point(), other.to_point() if isinstance(other, Position) else other).meters

    def to_point(self, with_altitude: bool = False) -> Point:
        """Convert to a ``geopy.point.Point``."""
        return Point(self.latitude, self.longitude, altitude=self.altitude if with_altitude else None)


class API(metaclass=ABCMeta):
    """A class for retrieving data from an API."""

    API_URL: ClassVar[str]
    """The base URL for the API."""

    _data: dict[str, any]

    def __init__(self) -> None:
        """Initialize a new ``API``."""
        self._data = dict()

    def load(self, key: str, default: any = None) -> any:  # noqa: F402
        """Load data from the cache.

        Args:
            key: The key to load.
            default: The default value to return if the key is not present.

        Returns:
            The data if present, else the default.
        """
        return self._data.get(key, default)

    def store(self, key: str, value: any) -> None:
        """Store data in the cache.

        Args:
            key: The key the data should be stored under.
            value: The data to store.
        """
        self._data[key] = value

    def __getitem__(self, item: str) -> any:
        return self._data[item]

    def __setitem__(self, key: str, value: any) -> None:
        self._data[key] = value

    def init(self) -> None:
        """Initialize the API connection."""
        pass


def store(name: str | Callable[..., T] = None) -> Callable[..., T]:
    """
    Decorator / decorator factory to apply to an ``API`` method
    to immediately store the return value of the decorated method
    as the key ``name`` or the method name if left out.
    """
    def decorator(method: Callable[..., T]) -> Callable[..., T]:
        @wraps(method)
        def wrapper(self: API, *args: any, **kwargs: any) -> T:
            if isinstance(self, API):
                self[name or method.__name__] = method(self, *args, **kwargs)
                return self[name or method.__name__]
            return method(self, *args, **kwargs)

        return wrapper

    if isinstance(name, str) or name is None:
        return decorator

    if callable(name):
        m = name
        name = m.__name__
        return decorator(m)

    raise ValueError('You need to apply this decorator to a method of an API!')
    # Really any callable works just fine, but in this case the decorator will do nothing


class ThreadedAPI(API, Thread):
    """An ``API`` that refreshes the data in a new thread."""

    _is_running: bool
    ready: Event

    def __init__(self) -> None:
        """Initialize a new ``ThreadedAPI``."""
        API.__init__(self)
        Thread.__init__(
            self,
            target=self._run,
            name=f"API-Thread for '{self.API_URL}'",
            daemon=True,
        )
        self._is_running = False
        self.ready = Event()

    @property
    def is_connected(self) -> bool:
        """Check whether the connector is connected to the server"""
        return self.ready.is_set() and self._is_running

    def _run(self) -> None:
        """The main loop that will run in a separate thread."""
        # thread join checks per second
        tps = 20
        counter = 0
        self._is_running = True
        while self._is_running:
            # If the counter is not 0, just wait and check for a thread join
            if counter != 0:
                time.sleep(1 / tps)
                counter = (counter + 1) % tps
                continue

            # The target time for when to perform the next refresh after this one
            target = time.time_ns() + int(1e9)

            try:
                self.refresh()
                if not self.ready.is_set():
                    self.ready.set()
            except (APIConnectionError, JSONDecodeError) as e:
                if not self.ready.is_set():
                    raise InitialConnectionError from e
                logger.exception(e)
                continue

            counter = (tps - int(max(0.0, (target - time.time_ns()) / 1e9) * tps)) % tps

    def stop(self) -> None:
        """Stop requesting data and shut down the separate thread."""
        self._is_running = False
        if self.is_alive():
            self.join()

    def reset(self) -> None:
        """Reset the thread and the cache so that they can be reused with ``start()``."""
        self.stop()
        self.__init__()

    @abstractmethod
    def refresh(self) -> None:
        """Method that collects data from the server and stores it in the cache."""
        raise NotImplementedError


class BlockingRestAPI(APISession, API):
    """A RESTful ``API`` that uses an ``restfly.session.APISession`` to fetch data."""

    _error_map = {
        501: APIFeatureMissingError,
    }

    def __init__(self, **kwargs: any) -> None:
        """Initialize a new ``BlockingRestAPI``.

        Args:
            kwargs: The kwargs to pass to the underlying ``restfly.session.APISession``.
        """
        kwargs['url'] = kwargs.pop('url', self.API_URL)
        APISession.__init__(self, **kwargs)
        API.__init__(self)

    def _build_session(self, **kwargs: any) -> None:
        APISession._build_session(self, **kwargs)
        self._session.headers.update({"User-Agent": "Python/onboardapis (%s)" % get_package_version()})


class ThreadedRestAPI(ThreadedAPI, BlockingRestAPI, metaclass=ABCMeta):
    """A threaded version of the ``BlockingRestAPI``."""

    def __init__(self, **kwargs: any) -> None:
        """Initialize a new ``ThreadedRestAPI``."""
        kwargs['url'] = kwargs.pop('url', self.API_URL)
        ThreadedAPI.__init__(self)
        BlockingRestAPI.__init__(self, **kwargs)


class BlockingGraphQlAPI(Client, API):
    """A GraphQL ``API`` that uses a ``gql.client.Client`` to fetch data."""

    def __init__(self, **kwargs: any) -> None:
        """Initialize a new ``BlockingGraphQlAPI``.

        Args:
            kwargs: The kwargs to pass to the underlying ``gql.client.Client``.
        """
        kwargs.update({
            'transport': kwargs.pop('transport', RequestsHTTPTransport(
                url=self.API_URL,
                headers={
                    'Content-Type': 'application/json',
                    'User-Agent': 'Python/onboardapis (%s)' % get_package_version(),
                },
            )),
            'fetch_schema_from_transport': kwargs.pop('fetch_schema_from_transport', True),
        })
        Client.__init__(self, **kwargs)
        API.__init__(self)


class ThreadedGraphQlAPI(ThreadedAPI, BlockingGraphQlAPI, metaclass=ABCMeta):
    """A threaded version of the ``BlockingGraphQlAPI``."""

    def __init__(self, **kwargs: any) -> None:
        """Initialize a new ``ThreadedGraphQlAPI``.

        Args:
            kwargs: The kwargs to pass to the underlying ``gql.client.Client``.
        """
        ThreadedAPI.__init__(self)
        BlockingGraphQlAPI.__init__(self, **kwargs)



"""
Package for third parties that provide API services and are used by other operators.

Contains the following third parties:

- ``onboardapis.train.third_party.icomera`` - Icomera AB
- ``onboardapis.train.third_party.unwired`` - Unwired Networks GmbH

---
"""
from __future__ import annotations

__all__ = [
    "icomera",
    "unwired",
]



from __future__ import annotations

import logging

from ...third_party.unwired import GenericUnwiredTrain
from ...third_party.unwired.mixins import UnwiredJourneyMixin

logger = logging.getLogger(__name__)


class SBahnHannover(GenericUnwiredTrain, UnwiredJourneyMixin):
    """Implementation for the S-Bahn Hannover."""

    def __init__(self) -> None:
        GenericUnwiredTrain.__init__(self)
        UnwiredJourneyMixin.__init__(self)



"""
Implementations for the ALSTOM Transportation Germany GmbH.

Contains the following APIs:

- ``MetronomCaptivePortal``

---
"""
from __future__ import annotations

from ..me import MetronomCaptivePortal

__all__ = [
    "MetronomCaptivePortal",
]



"""
Implementation for the German operator Transdev Hannover GmbH.

Contains the following APIs:

- ``SBahnHannover``

---
"""
from __future__ import annotations

from .apis import SBahnHannover

__all__ = [
    "SBahnHannover",
]



"""
Exceptions for the onboardapis package.
"""
from __future__ import annotations

from restfly.errors import APIError


class OnboardException(Exception):
    """Base exception class for on-board exceptions."""

    __slots__ = ()


class APIConnectionError(OnboardException, ConnectionError):
    """Exception raised when an API connection error occurs."""

    __slots__ = ()


class InitialConnectionError(APIConnectionError):
    """Exception raised only when the first try to connect to an API fails."""

    __slots__ = ()

    def __init__(self, message: str = "Unable to connect to the API, are you connected to the on-board Wi-Fi?", *args):
        super(InitialConnectionError, self).__init__(message, *args)


class DataInvalidError(APIConnectionError):
    """Error raised when an API sends a response that could not be parsed
    due to a formatting error or when the data is missing."""

    __slots__ = ()

    def __init__(self, message: str = "The API returned an invalid response", *args):
        super(DataInvalidError, self).__init__(message, *args)


class APIFeatureMissingError(APIConnectionError, APIError):
    """Error raised when an API could support a feature,
    but the actual implementation onboard this vehicle does not support this feature."""

    __slots__ = ()

    def __init__(self, resp, message: str = None, **kwargs):
        if not message:
            message = "The requested feature (%s) is not supported by this API" % resp.url
        APIConnectionError.__init__(self, message)
        APIError.__init__(self, resp, **kwargs)
        self.set_retryable(value=False)



from __future__ import annotations

import logging

from datetime import datetime, timedelta
from typing import Generator

from ... import ConnectingTrain
from ....data import ThreadedRestAPI, ScheduledEvent, store, ID

logger = logging.getLogger(__name__)


class InouiConnector(ThreadedRestAPI):
    API_URL = "https://wifi.sncf/router/api"

    @store('gps')
    def gps(self) -> dict:
        return self.get("train/gps").json()

    @store('details')
    def details(self) -> dict:
        return self.get("train/details").json()

    def trainboard(self, station_id: ID) -> dict:
        return self.get(f"information/trainboard/{station_id}")

    def refresh(self) -> None:
        self.gps()
        self.details()

        # self._api["attendance"] = self.get("bar/attendance").json()
        # self._api["auto"] = self.get("connection/activate/auto").json()
        # self._api["catalog"] = self.get("food/catalog").json()
        # self._api["coverage"] = self.get("train/coverage").json()
        # self._api["graph"] = self.get("train/graph").json()
        # self._api["modules"] = self.get("configuration/modules").json()
        # self._api["poi"] = self.get("poi").json()
        # self._api["registry"] = self.get("connection/registry").json()
        # self._api["room"] = self.get("chat/room").json()
        # self._api["statistics"] = self.get("connection/statistics").json()
        # self._api["status"] = self.get("connection/status").json()
        # self._api["videos"] = self.get("media/videos").json()
        # self._api["wordings"] = self.get("media/wordings", params=dict(language="de")).json()

    def connections(self, station_id: ID) -> Generator[ConnectingTrain, None, None]:
        for connection in self.trainboard(station_id).get("train", []):
            yield ConnectingTrain(
                vehicle_type=connection.get("type"),
                line_number=connection.get("num"),
                departure=ScheduledEvent(
                    scheduled=datetime.fromisoformat(connection.get("heure")),
                    actual=datetime.fromisoformat(connection.get("heure"))
                    + timedelta(minutes=int(connection.get("retard", 0))),
                ),
                destination=connection.get("origdest"),
                platform=ScheduledEvent(
                    scheduled=connection.get("attribut_voie"),
                    actual=connection.get("attribut_voie"),
                ),
            )



"""
Package for buses

## Operator ID

The operator ID for buses is highly dependent on the region
and will be described further in the according country package.
"""
from __future__ import annotations

from abc import ABCMeta

from .. import Vehicle

__all__ = [
    "Bus",
]


class Bus(Vehicle, metaclass=ABCMeta):
    """
    Base class for buses
    """



from __future__ import annotations

import logging

from ....units import meters_per_second
from ....data import Position
from ....mixins import PositionMixin, SpeedMixin
from ... import Train
from .interfaces import FlixTainmentAPI

logger = logging.getLogger(__name__)


class FlixTainment(Train, SpeedMixin, PositionMixin):
    """Wrapper for interacting with the Flixtrain FLIXTainment API
    (few methods are available, because the API is very sparse)."""

    _api: FlixTainmentAPI

    def __init__(self):
        self._api = FlixTainmentAPI()
        Train.__init__(self)

    @property
    def position(self) -> Position:
        return Position(
            latitude=self._api["position"].get("latitude", None),
            longitude=self._api["position"].get("longitude", None),
        )

    @property
    def speed(self) -> float:
        return meters_per_second(kilometers_per_hour=float(self._api["position"].get("speed", 0.0)))

    @property
    def type(self) -> str:
        return 'FLX'



from __future__ import annotations

import logging

from ....data import ThreadedRestAPI, store

logger = logging.getLogger(__name__)


class FlixTainmentAPI(ThreadedRestAPI):
    API_URL = "https://media.flixtrain.com"

    @store('position')
    def position(self) -> dict:
        return self.get("services/pis/v1/position")

    def refresh(self) -> None:
        self.position()



from __future__ import annotations

import logging
import re

from ....mixins import InternetAccessMixin
from ....data import ID
from ... import Train
from .interfaces import GenericUnwiredAPI, GenericUnwiredInternetAccessInterface

logger = logging.getLogger(__name__)


class GenericUnwiredTrain(Train, InternetAccessMixin):
    _api: GenericUnwiredAPI
    _internet_access: GenericUnwiredInternetAccessInterface

    def __init__(self):
        if not hasattr(self, '_api'):
            self._api = GenericUnwiredAPI()
        if not hasattr(self, '_internet_access'):
            self._internet_access = GenericUnwiredInternetAccessInterface(self._api)
        Train.__init__(self)
        InternetAccessMixin.__init__(self)

    @property
    def id(self) -> ID:
        return self._api['journey']['id']

    @property
    def type(self) -> str:
        return re.match(r'(\w+)\d+', self._api['journey']['line']).group(1)

    @property
    def line_number(self) -> str:
        return re.match(r'\w+(\d+)', self._api['journey']['line']).group(1)



from __future__ import annotations

import logging

from ...third_party.icomera import GenericIcomeraTrain
from .interfaces import MetronomAPI, MetronomInternetAccessInterface

logger = logging.getLogger(__name__)


class MetronomCaptivePortal(GenericIcomeraTrain):
    _api: MetronomAPI
    _internet_access: MetronomInternetAccessInterface

    def __init__(self) -> None:
        self._api = MetronomAPI()
        self._internet_access = MetronomInternetAccessInterface(self._api)
        GenericIcomeraTrain.__init__(self)

    @property
    def type(self) -> str:
        return 'ME'



from __future__ import annotations

import logging

from ....data import ThreadedRestAPI, store

logger = logging.getLogger(__name__)


class PortaleRegionaleConnector(ThreadedRestAPI):
    API_URL = 'https://www.portaleregionale.it/PortaleRegionale'

    @store('infovaggio')
    def infovaggio(self) -> dict:
        return self.get(
            "infoviaggio.getData.action",
            params={"stazioniList": True},
        ).json()

    @store('m53')
    def m53(self) -> dict:
        return self.get("m53.getData.action").json()

    @store('map')
    def map(self) -> dict:
        return self.get("map.getData.action").json()

    @store('meteo')
    def meteo(self) -> dict:
        return self.get("meteo.getData.action").json()

    @store('stations')
    def stations(self) -> dict:
        return self.get("stations.getData.action").json()

    @store('common')
    def common(self) -> dict:
        return self.get("common.getInfos.action").json()

    def refresh(self):
        self.infovaggio()
        # self.m53()
        # self.map()
        # self.meteo()
        # self.stations()
        # self.common()



"""
Package for ships

## Operator ID

The operator ID for ships is the <undefined>.
"""
from __future__ import annotations

from abc import ABCMeta

from .. import Vehicle

__all__ = [
    "Ship",
]


class Ship(Vehicle, metaclass=ABCMeta):
    """
    Base class for ships
    """



"""
Package for German trains.

Contains the following operators:

- ``onboardapis.train.de.bth`` - ALSTOM Transportation Germany GmbH
- ``onboardapis.train.de.db`` - Deutsche Bahn / DB AG
- ``onboardapis.train.de.flix`` - Flix Train GmbH
- ``onboardapis.train.de.me`` - metronom Eisenbahngesellschaft mbH
- ``onboardapis.train.de.tdh`` - Transdev Hannover GmbH

---
"""
from __future__ import annotations

__all__ = [
    "bth",
    "db",
    "flix",
    "me",
    "tdh",
]



"""
Implementations for Unwired Networks GmbH.

Contains the following APIs:

- ``GenericUnwiredTrain``

---
"""
from __future__ import annotations

from .apis import GenericUnwiredTrain
from .interfaces import GenericUnwiredAPI, GenericUnwiredInternetAccessInterface
from .mixins import UnwiredJourneyMixin, UnwiredMapMixin

__all__ = [
    "GenericUnwiredTrain",
    "GenericUnwiredAPI",
    "GenericUnwiredInternetAccessInterface",
    "UnwiredJourneyMixin",
    "UnwiredMapMixin",
]



"""
Implementation of the Austrian operator Österreichische Bundesbahnen.

Contains the following APIs:

- ``RailnetRegio``

---
"""
from __future__ import annotations

from .apis import RailnetRegio

__all__ = [
    "RailnetRegio",
]



from __future__ import annotations

import logging

from datetime import datetime, timedelta

from ....data import ID
from ....exceptions import DataInvalidError
from ....units import meters_per_second
from ... import Train, TrainStation
from .interfaces import PortaleRegionaleConnector

logger = logging.getLogger(__name__)


class PortaleRegionale(Train):
    """
    Wrapper for interacting with the Trenitalia PortaleRegionale API
    """

    _api: PortaleRegionaleConnector

    _stations: dict[ID, TrainStation]
    """A dict that contains the known stations (origin, destination and passed stations)"""

    def __init__(self):
        self._api = PortaleRegionaleConnector()
        self._stations = dict()
        Train.__init__(self)

    @property
    def now(self) -> datetime:
        return datetime.fromisoformat(self._api['infovaggio']['datetime'])

    @property
    def id(self) -> ID:
        return self.line_number

    @property
    def line_number(self) -> str:
        return self._api['infovaggio']['infos']['trackNumber']

    def delay(self) -> timedelta:
        delay_minutes = self._api['infovaggio']['infos'].get('delay', '')
        return timedelta() if delay_minutes in ('0', '') else timedelta(minutes=int(delay_minutes))

    @property
    def speed(self) -> float:
        if self._api['infovaggio'].get('isGpsValid', 'false').lower() == 'true':
            return meters_per_second(kilometers_per_hour=(float(self._api['infovaggio']['infos']['speed'])))
        raise DataInvalidError('GPS data invalid.')

    @property
    def stations_dict(self) -> dict[ID, TrainStation]:
        first = self._api['infovaggio']['infos']['stazionePartenza']
        current = self._api['infovaggio']['nextStation']
        last = self._api['infovaggio']['infos']['stazioneArrivo']

        if len(self._stations) == 0:
            self._stations = {
                first: TrainStation(
                    id=first, name=first,
                    arrival=None, departure=None, position=None, distance=None, _connections=(), platform=None
                ),
                current: TrainStation(
                    id=current, name=current,
                    arrival=None, departure=None, position=None, distance=None, _connections=(), platform=None
                ),  # if current == first Python will just overwrite the entry
                last: TrainStation(
                    id=last, name=last,
                    arrival=None, departure=None, position=None, distance=None, _connections=(), platform=None
                ),
            }
            return self._stations.copy()

        self._stations = {
            first: TrainStation(
                id=first, name=first,
                arrival=None, departure=None, position=None, distance=None, _connections=(), platform=None
            ),
            **{key: value for key, value in tuple(self._stations.items())[1:-1]},
            current: TrainStation(
                id=current, name=current,
                arrival=None, departure=None, position=None, distance=None, _connections=(), platform=None
            ),
            last: TrainStation(
                id=last, name=last,
                arrival=None, departure=None, position=None, distance=None, _connections=(), platform=None
            ),  # if current == last Python will just overwrite the entry
        }
        return self._stations.copy()

    @property
    def current_station(self) -> TrainStation:
        return self.stations_dict[self._api['infovaggio']['nextStation']]



from __future__ import annotations

import logging
from abc import ABCMeta
from datetime import datetime, timedelta

from ....mixins import SpeedMixin, PositionMixin, StationsMixin
from ....data import Position, ScheduledEvent, ID
from ... import Train, TrainStation

logger = logging.getLogger(__name__)

__all__ = [
    "UnwiredJourneyMixin",
    "UnwiredMapMixin",
]


class UnwiredJourneyMixin(Train, StationsMixin[TrainStation], metaclass=ABCMeta):
    @property
    def stations_dict(self) -> dict[ID, TrainStation]:
        return {
            station['id']: TrainStation(
                id=station['id'],
                name=station['name'],
                arrival=ScheduledEvent(
                    scheduled=(
                        datetime.fromisoformat(station['arrivalPlanned']).replace(tzinfo=None)
                        if 'arrivalPlanned' in station.keys() else None
                    ),
                    actual=(
                        datetime.fromisoformat(station['arrivalPlanned']).replace(tzinfo=None)
                        + timedelta(minutes=station['arrivalDelay'])
                        if 'arrivalDelay' in station.keys() else None
                    )
                ),
                departure=ScheduledEvent(
                    scheduled=(
                        datetime.fromisoformat(station['departurePlanned']).replace(tzinfo=None)
                        if 'departurePlanned' in station.keys() else None
                    ),
                    actual=(
                        datetime.fromisoformat(station['departurePlanned']).replace(tzinfo=None)
                        + timedelta(minutes=station['departureDelay'])
                        if 'departureDelay' in station.keys() else None
                    ),
                ),
                distance=None,
                position=None,
                platform=None,
                _connections=(),
            )
            for station in self._api['journey']['stops']
        }

    @property
    def current_station(self) -> TrainStation:
        station, *_ = filter(
            lambda s: self.now < s.arrival.actual,
            filter(
                lambda s: s.arrival.actual is not None,
                self.stations
            )
        )
        return station


class UnwiredMapMixin(SpeedMixin, PositionMixin, metaclass=ABCMeta):
    @property
    def position(self) -> Position:
        raise NotImplementedError



"""
.. include:: ../README.md

---
"""
from __future__ import annotations

import logging

from abc import ABCMeta
from dataclasses import dataclass
from datetime import datetime
from typing import Iterable

from .data import ID, API, ThreadedAPI, ScheduledEvent, Position
from .exceptions import InitialConnectionError

logger = logging.getLogger(__name__)

__all__ = [
    "bus",
    "data",
    "exceptions",
    "mixins",
    "other",
    "plane",
    "protocols",
    "ship",
    "train",
    "units",
    "Vehicle",
    "Station",
    "ConnectingVehicle",
]


class Vehicle(metaclass=ABCMeta):
    """
    Base class for all vehicles
    """

    _api: API
    """The :class:`API` that supplies the data for this vehicle."""

    def __enter__(self):
        self.init()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()

    def init(self) -> None:
        """Initialize the connection to the API.

        Call the init method of the API that supplies the data for this vehicle.

        Raises:
          InitialConnectionError: If the connection to the API could not be established.
        """
        if not hasattr(self, '_api'):
            return  # Abstract class without API implementation

        try:
            self._api.init()
            if isinstance(self._api, ThreadedAPI):
                self._api.start()
                if not self._api.ready.wait(timeout=15):
                    raise RuntimeError
                return
        except RuntimeError as e:
            raise InitialConnectionError from e

    def shutdown(self) -> None:
        """
        This method will be called when exiting the context manager.

        :return: Nothing
        :rtype: None
        """
        if not hasattr(self, '_api'):
            return  # Abstract class without API implementation

        if isinstance(self._api, ThreadedAPI):
            self._api.stop()

    @property
    def now(self) -> datetime:
        """
        Get the current time as seen by the vehicle

        :return: The current time
        :rtype: datetime.datetime
        """
        return datetime.now()

    @property
    def id(self) -> ID:
        """The unique ID of this specific vehicle."""
        return 'undefined'


@dataclass
class ConnectingVehicle(object):
    """
    A connecting vehicle is a vehicle that is not part of the main trip but of a connecting service.
    It may only have limited information available.
    """
    vehicle_type: str | None
    """The abbreviated vehicle type"""
    line_number: str | None
    """The line number of the vehicle"""
    departure: ScheduledEvent[datetime] | None
    """The departure time of the vehicle"""
    destination: str | None
    """The destination of the vehicle"""


@dataclass
class Station(object):
    """
    A Station is a stop on the trip
    """
    id: ID
    """The ID of the station"""
    name: str
    """The name of the station"""
    arrival: ScheduledEvent[datetime] | None
    """The arrival time at this station"""
    departure: ScheduledEvent[datetime] | None
    """The departure time from this station"""
    position: Position | None
    """The geographic position of the station"""
    distance: float | None
    """The distance from the start to this station"""
    _connections: Iterable[ConnectingVehicle]

    def __str__(self) -> str:
        return self.name

    @property
    def connections(self) -> list[ConnectingVehicle]:
        """The connecting services departing from this station."""
        if not isinstance(self._connections, list):
            self._connections = list(self._connections)
        return self._connections

    def calculate_distance(self, other: Station | Position | int | float) -> float | None:
        """
        Calculate the distance in meters between this station and something else.

        Accepts a :class:`Station`, :class:`Position` or a number for the distance calculation.

        :param other: The other station or position to calculate the distance to
        :type other: Station | Position | int | float
        :return: The distance in meters
        :rtype: Optional[float]
        """
        # If there is not enough information to calculate the distance, return None
        if other is None:
            return None

        # Both distances since the start are known
        if isinstance(other, (int, float)) and self.distance is not None:
            return abs(self.distance - other)

        # Both positions are known
        if isinstance(other, Position) and self.position is not None:
            return self.position.calculate_distance(other)

        # Both are a station
        if isinstance(other, Station):
            if self.distance is not None and other.distance is not None:
                return abs(self.distance - other.distance)
            if self.position is not None and other.position is not None:
                return self.position.calculate_distance(other.position)



from __future__ import annotations

import logging

from datetime import datetime
from enum import Enum
from functools import lru_cache
from json import JSONDecodeError
from typing import Generator
from urllib.parse import urlparse, parse_qs

from bs4 import BeautifulSoup
from geopy import Point
from geopy.distance import distance
from restfly.errors import NotFoundError

from ....data import (
    ID,
    ThreadedRestAPI,
    ScheduledEvent,
    default,
    store,
    BlockingRestAPI, get_package_version,
)
from ....exceptions import APIFeatureMissingError
from ....mixins import InternetAccessInterface, InternetMetricsInterface
from ... import ConnectingTrain

logger = logging.getLogger(__name__)


class ICEPortalAPI(ThreadedRestAPI):
    API_URL = "https://iceportal.de"

    @store('bap')
    @lru_cache
    def bap_service_status(self) -> dict[str, any]:
        try:
            return self.get("bap/api/bap-service-status").json()
        except (JSONDecodeError, APIFeatureMissingError):
            return {'status': 'false'}

    @store('trip')
    def trip_info(self) -> dict[str, any]:
        return self.get("api1/rs/tripInfo/trip").json()

    @store('status')
    def status(self) -> dict[str, any]:
        return self.get("api1/rs/status").json()

    def refresh(self) -> None:
        self.status()
        self.trip_info()
        self.bap_service_status()

    def get_connections(
            self, station_id: str
    ) -> Generator[ConnectingTrain, None, None]:
        """
        Get all connections for a station

        :param station_id: The station to get the connections for
        :type station_id: str
        :return: A generator yielding a list of connections for the station
        :rtype: Generator[list[ConnectingTrain]]
        """
        # Process the connections
        connections = list(
            [
                ConnectingTrain(
                    vehicle_type=connection.get("trainType", None),
                    line_number=connection.get("vzn", None),
                    platform=ScheduledEvent(
                        scheduled=connection.get("track", {}).get("scheduled", None),
                        actual=connection.get("track", {}).get("actual", None),
                    ),
                    destination=connection.get("station", {}).get("name", None),
                    departure=ScheduledEvent(
                        scheduled=(
                            datetime.fromtimestamp(int(default(
                                connection.get("timetable", {}).get("scheduledDepartureTime"), 0,
                            )) / 1000)
                            if default(connection.get("timetable", {}).get("scheduledDepartureTime")) is not None
                            else None
                        ),
                        actual=(
                            datetime.fromtimestamp(int(default(
                                connection.get("timetable", {}).get("actualDepartureTime"), 0
                            )) / 1000)
                            if default(connection.get("timetable", {}).get("actualDepartureTime")) is not None
                            else None
                        ),
                    ),
                )
                for connection in self.get(f"/api1/rs/tripInfo/connection/{station_id}").json().get("connections", [])
            ]
        )
        if len(connections) == 0:  # no data available
            cached = self._data.get(f"connections_{station_id}", [])
            yield from cached
            return

        self[f"connections_{station_id}"] = connections
        yield from connections
        return


class ICEPortalInternetAccessAPI(BlockingRestAPI):
    API_URL = 'https://login.wifionice.de'

    def __init__(self, **kwargs: any):
        super().__init__(**kwargs)

    def _build_session(self, **kwargs: any) -> None:
        super()._build_session(**kwargs)
        self._session.headers.update({'X-Requested-With': 'Python/onboardapis (%s)' % get_package_version()})


class ICEPortalInternetInterface(InternetAccessInterface, InternetMetricsInterface):
    _api: ICEPortalInternetAccessAPI

    def enable(self) -> None:
        logger_before = logging.getLogger('restfly.errors.NotFoundError').disabled
        logging.getLogger('restfly.errors.NotFoundError').disabled = True
        try:
            self._api.post('cna/logon', json={}, headers={
                'X-Csrf-Token': 'csrf',
            })
        except NotFoundError:
            # old login API
            response = self._api.get('de')
            self._api.post('de/', data={
                'login': True,
                'CSRFToken': response.cookies['csrf'],
            })
        finally:
            logging.getLogger('restfly.errors.NotFoundError').disabled = logger_before

    def disable(self):
        logger_before = logging.getLogger('restfly.errors.NotFoundError').disabled
        logging.getLogger('restfly.errors.NotFoundError').disabled = True
        try:
            self._api.post('cna/logoff', json={}, headers={'X-Csrf-Token': 'csrf'})  # Not CSRF protected anymore?
        except NotFoundError:
            response = self._api.get('de/')
            self._api.post('de/', data={
                'logout': True,
                'CSRFToken': response.cookies['csrf'],
            })
        finally:
            logging.getLogger('restfly.errors.NotFoundError').disabled = logger_before

    @property
    def is_enabled(self) -> bool:
        logger_before = logging.getLogger('restfly.errors.NotFoundError').disabled
        logging.getLogger('restfly.errors.NotFoundError').disabled = True
        try:
            return self._api.get('cna/wifi/user_info').json()['result']['authenticated'] == '1'
        except NotFoundError:
            return BeautifulSoup(
                self._api.get('de').text, 'html.parser'
            ).find(id='accept') is None
        finally:
            logging.getLogger('restfly.errors.NotFoundError').disabled = logger_before

    @property
    def limit(self) -> float | None:
        try:
            usage_info = self._api.get('usage_info/')
        except APIFeatureMissingError:
            return None
        return usage_info.json()['limit']


class ModeOfTransport(str, Enum):
    UNKNOWN = 'UNKNOWN'
    BUS = 'BUS'
    TRAM = 'TRAM'
    SUBWAY = 'SUBWAY'
    CITY_TRAIN = 'CITY_TRAIN'
    REGIONAL_TRAIN = 'REGIONAL_TRAIN'
    INTER_REGIONAL_TRAIN = 'INTER_REGIONAL_TRAIN'
    HIGH_SPEED_TRAIN = 'HIGH_SPEED_TRAIN'

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    def local_trains(cls) -> tuple[ModeOfTransport, ...]:
        return cls.CITY_TRAIN, cls.TRAM, cls.SUBWAY

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    def long_distance_trains(cls) -> tuple[ModeOfTransport, ...]:
        return cls.HIGH_SPEED_TRAIN, cls.REGIONAL_TRAIN, cls.INTER_REGIONAL_TRAIN

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    def trains(cls) -> tuple[ModeOfTransport, ...]:
        # noinspection PyUnresolvedReferences
        return cls.local_trains + cls.long_distance_trains


class RegioGuideAPI(ThreadedRestAPI):
    API_URL = "https://zugportal.de"

    @store('journey')
    def journey(self) -> dict:
        return self.get("@prd/zupo-travel-information/api/public/ri/journey").json()

    def refresh(self) -> None:
        self.journey()

    @lru_cache
    def distance(self, index: int) -> float:
        """Calculate the distance from the start for the station at index ``index``"""
        if index == 0:
            return 0.0

        start = self._data['journey'].get('stops', [])[0]
        stop = self._data['journey'].get('stops', [])[index]

        return self.distance(index - 1) + distance(Point(
            latitude=start.get('station', {}).get('position', {}).get('latitude'),
            longitude=start.get('station', {}).get('position', {}).get('longitude')
        ), Point(
            latitude=stop.get('station', {}).get('position', {}).get('latitude'),
            longitude=stop.get('station', {}).get('position', {}).get('longitude')
        )).meters

    def connections(self, station_id: ID) -> Generator[ConnectingTrain, None, None]:
        # noinspection PyTypeChecker
        yield from (
            ConnectingTrain(
                departure=ScheduledEvent(
                    scheduled=datetime.fromisoformat(item['timePredicted']),
                    actual=datetime.fromisoformat(item['time'])
                ),
                destination=item['station']['name'],
                line_number=item['train']['lineName'],
                platform=ScheduledEvent(
                    scheduled=item['platformPredicted'],
                    actual=item['platform'],
                ),
                vehicle_type=item['train']['category'],
            )
            for item
            in self.get(
                f'@prd/zupo-travel-information/api/public/ri/board/departure/{station_id}',
                params=dict(
                    modeOfTransport=','.join(map(lambda m: m.value, ModeOfTransport.trains)),
                    occupancy=True
                )
            ).json().get('items', [])
        )


ZugPortalAPI = RegioGuideAPI
"""Renamed by DB. Use RegioGuideAPI instead."""


class RegioGuideInternetAccessAPI(BlockingRestAPI):
    API_URL = 'http://192.168.44.1/'  # Hotsplots, covers every provider of RegioGuide? Same IP every time?

    def __init__(self, **kwargs: any):
        super().__init__(**kwargs)


class RegioGuideInternetAccessInterface(InternetAccessInterface):
    _api: RegioGuideInternetAccessAPI

    def enable(self) -> None:
        """WIP: does not work yet"""
        response = self._api.get('auth/login.php')
        params = parse_qs(urlparse(response.url).query)
        self._api.post('auth/login.php', data={
            "haveTerms": "1",
            "termsOK": "on",
            "button": "Jetzt kostenlos surfen",
            "challenge": params['challenge'][0],
            "uamip": params['uamip'][0],
            "uamport": params['uamport'][0],
            "userurl": params['userurl'][0],
            "myLogin": "agb",
            "ll": "de",
            "nasid": params['nasid'][0],
            "custom": "1",
        })

    def disable(self) -> None:
        """WIP: does not work yet"""
        response = self._api.get('/logoff')  # Not CSRF protected!

    @property
    def is_enabled(self) -> bool:
        """WIP: does not work reliably yet"""
        response = self._api.get('/auth/login.php')
        soup = BeautifulSoup(response.text, 'html.parser')
        return soup.select_one('[href*="/logoff"]') is not None



from __future__ import annotations

import logging

from ...third_party.icomera import GenericIcomeraAPI, GenericIcomeraInternetAccessInterface

logger = logging.getLogger(__name__)


class MetronomAPI(GenericIcomeraAPI):
    # noinspection HttpUrlsUsage
    API_URL = 'http://wifi.metronom.de'


class MetronomInternetAccessInterface(GenericIcomeraInternetAccessInterface):
    # noinspection HttpUrlsUsage
    API_URL = 'http://wifi.metronom.de'



"""
Implementation for the French train operator SNCF Voyageurs.

Contains the following APIs:

- ``PortalINOUI``

---
"""
from __future__ import annotations

from .apis import PortalINOUI

__all__ = [
    "PortalINOUI",
]



from __future__ import annotations

import logging
import re

from typing import Literal
from datetime import datetime, timedelta

from deprecation import deprecated

from ....exceptions import DataInvalidError
from ....data import ID, default, ScheduledEvent, Position, get_package_version
from ....mixins import SpeedMixin, PositionMixin, StationsMixin, InternetAccessMixin
from ....units import meters_per_second
from ... import Train, TrainStation
from .mappings import id_name_map
from .interfaces import (
    ICEPortalAPI,
    ICEPortalInternetInterface,
    ICEPortalInternetAccessAPI,
    RegioGuideAPI,
    RegioGuideInternetAccessInterface,
    RegioGuideInternetAccessAPI,
)

logger = logging.getLogger(__name__)

InternetStatus = Literal["NO_INFO", "NO_INTERNET", "UNSTABLE", "WEAK", "MIDDLE", "HIGH"]


class ICEPortal(Train, SpeedMixin, PositionMixin, StationsMixin[TrainStation], InternetAccessMixin):
    """Wrapper for interacting with the DB ICE Portal API."""

    _api: ICEPortalAPI
    _internet_access: ICEPortalInternetInterface

    def __init__(self):
        self._api = ICEPortalAPI()
        self._internet_access = ICEPortalInternetInterface(ICEPortalInternetAccessAPI())
        Train.__init__(self)

    @property
    def now(self) -> datetime:
        return datetime.fromtimestamp(int(default(
            self._api["status"].get("serverTime", None), 0,
        )) / 1000)

    @property
    def id(self) -> str:
        return self._api["status"].get("tzn")

    @property
    def type(self) -> str:
        return self._api["trip"].get("trip", {}).get("trainType")

    @property
    def line_number(self) -> str:
        return self._api["trip"].get("trip", {}).get("vzn")

    @property
    def stations_dict(self) -> dict[str, TrainStation]:
        stops = self._api["trip"].get("trip", {}).get("stops")
        if stops is None:
            raise DataInvalidError("API is missing data about stations")
        return {
            stop.get("station", {}).get("evaNr"): TrainStation(
                id=stop.get("station", {}).get("evaNr"),
                name=stop.get("station", {}).get("name"),
                platform=ScheduledEvent(
                    scheduled=stop.get("track", {}).get("scheduled"),
                    actual=stop.get("track", {}).get("actual"),
                ),
                arrival=ScheduledEvent(
                    scheduled=datetime.fromtimestamp(int(default(
                        stop.get("timetable", {}).get("scheduledArrivalTime"), 0,
                    )) / 1000)
                    if default(stop.get("timetable", {}).get("scheduledArrivalTime")) is not None
                    else None,
                    actual=datetime.fromtimestamp((default(
                        stop.get("timetable", {}).get("actualArrivalTime"), 0,
                    )) / 1000)
                    if default(stop.get("timetable", {}).get("actualArrivalTime")) is not None
                    else None,
                ),
                departure=ScheduledEvent(
                    scheduled=datetime.fromtimestamp(int(default(
                        stop.get("timetable", {}).get("scheduledDepartureTime"), 0,
                    )) / 1000)
                    if default(stop.get("timetable", {}).get("scheduledDepartureTime")) is not None
                    else None,
                    actual=datetime.fromtimestamp(int(default(
                        stop.get("timetable", {}).get("actualDepartureTime"), 0,
                    )) / 1000)
                    if default(stop.get("timetable", {}).get("actualDepartureTime")) is not None
                    else None,
                ),
                position=Position(
                    latitude=stop.get("station", {}).get("geocoordinates", {}).get("latitude"),
                    longitude=stop.get("station", {}).get("geocoordinates", {}).get("longitude"),
                ),
                distance=stop.get("info", {}).get("distanceFromStart", 0),
                _connections=self._api.get_connections(station_id=stop.get("station", {}).get("evaNr")),
            )
            for stop in stops
        }

    @property
    def current_station(self) -> TrainStation:
        # Get the current station id
        stop_info = (
            self._api["trip"].get("trip", {}).get("stopInfo", {})
        )
        station_id = default(stop_info.get("actualNext"))
        if station_id is None:  # None if the arrival time of the last station has passed
            stop, *_ = *filter(
                lambda s: not s.get("info", {}).get("passed", True),
                self._api["trip"].get("trip", {}).get("stops", [])
            ), None
            if stop is None:  # None if all stations have been passed
                return self.destination
            station_id = stop.get("station", {}).get("evaNr")
        # Get the station from the stations dict
        try:
            return self.stations_dict[station_id]
        except KeyError as e:
            raise DataInvalidError("No current station found") from e

    @property
    def speed(self) -> float:
        return meters_per_second(kilometers_per_hour=self._api["status"].get("speed", 0))

    @property
    def distance(self) -> float:
        return (
            self._api["trip"].get("trip", {}).get("actualPosition", 0)
            + self._api["trip"].get("trip", {}).get("distanceFromLastStop", 0)
        )

    @property
    def position(self) -> Position:
        return Position(
            latitude=self._api["status"].get("latitude"),
            longitude=self._api["status"].get("longitude"),
        )

    @property
    def name(self) -> str | None:
        """
        Get the name of the train.

        Most of the DB ICE train have names.
        Names are not available through the API, instead a public list of names will be used.

        :return: The name of the train
        """
        match = re.search(r"\d+", f'{self.id}')
        if match is None:
            return None

        return id_name_map.get(int(match.group(0)))

    @property
    def all_delay_reasons(self) -> dict[str, list[str]]:
        """
        Get all delay reasons for the current trip

        :return: A dictionary of delay reasons with the station id as the key
        :rtype: dict[str, list[str] | None]
        """
        return {
            stop.get("station", {}).get("evaNr", None): list(
                [
                    default(reason.get("text", None))
                    for reason in default(stop.get("delayReasons"), [])
                ]
            )
            for stop in self._api["trip"]
            .get("trip", {})
            .get("stops", [])
        }

    @property
    def delay_reasons(self) -> list[str] | None:
        """
        Get the delay reason for the current station

        :return: The delay reason
        :rtype: list[str] | None
        """
        return self.all_delay_reasons.get(self.current_station.id, None)

    @property
    def has_bap(self) -> bool:
        """
        Returns True if the train has a BAP module installed and active

        bap = 'Bestellen am Platz' is a service that allows passengers to order food and drinks right to their seat
        """
        # bap is a service exclusive to first class
        if self.wagon_class != "FIRST":
            return False
        # Check if the module is installed
        if str(self._api["status"].get("bapInstalled", False)).lower() != "true":
            return False
        # Check if the module is active
        return str(self._api["bap"].get("status", False)).lower() == "true"

    @property
    def wagon_class(self) -> Literal["FIRST", "SECOND"] | None:
        """
        Get the wagon class of the wagon you are currently in

        :return: The wagon class
        :rtype: Literal["FIRST", "SECOND"]
        """
        return default(self._api["status"].get("wagonClass"))

    @property
    def internet_status(self) -> InternetStatus:
        """Returns the current internet connection status of the train."""
        return default(
            self._api["status"]
            .get("connectivity", {})
            .get("currentState"),
            "NO_INFO",
        )

    @property
    def next_internet_status(self) -> InternetStatus:
        """Returns the next internet connection status of the train."""
        return default(
            self._api["status"]
            .get("connectivity", {})
            .get("nextState"),
            "NO_INFO",
        )

    @property
    def internet_status_change(self) -> timedelta | None:
        """Returns the time until ``internet_status`` changes to ``next_internet_status``."""
        remaining_seconds = (
            self._api["status"]
            .get("connectivity", {})
            .get("remainingTimeSeconds", "")
        )
        return None if default(remaining_seconds) is None else timedelta(seconds=int(remaining_seconds))


class RegioGuide(Train, StationsMixin[TrainStation], InternetAccessMixin):
    """Wrapper for interacting with the DB Regio-Guide API, formerly Zug Portal."""

    _api: RegioGuideAPI
    _internet_access: RegioGuideInternetAccessInterface

    def __init__(self):
        self._api = RegioGuideAPI()
        self._internet_access = RegioGuideInternetAccessInterface(RegioGuideInternetAccessAPI())
        Train.__init__(self)

    @property
    def now(self) -> datetime:
        return datetime.fromisoformat(self._api['journey'].get('date'))

    @property
    def id(self) -> ID:
        return self._api['journey'].get('no')

    @property
    def type(self) -> str:
        return self._api['journey'].get('category')

    @property
    def line_number(self) -> str:
        return self._api['journey'].get('name').lstrip(self.type).strip()

    @property
    def stations_dict(self) -> dict[ID, TrainStation]:
        return {
            stop.get('station', {}).get('evaNo'): TrainStation(
                id=stop.get('station', {}).get('evaNo'),
                name=stop.get('station', {}).get('name'),
                platform=ScheduledEvent(
                    scheduled=stop.get('track', {}).get('target'),
                    actual=stop.get('track', {}).get('prediction')
                ),
                arrival=None if stop.get('arrivalTime') is None else ScheduledEvent(
                    scheduled=datetime.fromisoformat(stop.get('arrivalTime', {})['target']),
                    actual=datetime.fromisoformat(stop.get('arrivalTime', {})['predicted'])
                ),
                departure=None if stop.get('departureTime') is None else ScheduledEvent(
                    scheduled=datetime.fromisoformat(stop.get('departureTime', {})['target']),
                    actual=datetime.fromisoformat(stop.get('departureTime', {})['predicted'])
                ),
                position=Position(
                    latitude=stop.get('station', {}).get('position', {}).get('latitude'),
                    longitude=stop.get('station', {}).get('position', {}).get('longitude')
                ),
                distance=self._api.distance(index),
                _connections=self._api.connections(stop.get('station', {}).get('evaNo')),
            )
            for index, stop in enumerate(self._api['journey'].get('stops', []))
        }

    @property
    def current_station(self) -> TrainStation:
        station, *_ = *filter(
            lambda s: self.now < s.arrival.actual, filter(lambda s: s.arrival is not None, self.stations)
        ), None
        return self.destination if station is None else station


@deprecated(
    deprecated_in='2.0.0rc3',
    removed_in='2.0.0',
    current_version=get_package_version(),
    details='Renamed by DB. Use RegioGuide instead.'
)
class ZugPortal(RegioGuide):
    """Renamed by DB. Use RegioGuide instead."""



from __future__ import annotations

id_name_map: dict[int, str | None] = {
    1: "Gießen",
    3: "Neu-Isenburg",
    5: "Offenbach am Main",
    6: "Itzehoe",
    7: "Plattling",
    10: "Gelsenkirchen",
    11: "Nürnberg",
    12: "Memmingen",
    13: "Frankenthal/Pfalz",
    14: "Friedrichshafen",
    # 15: "Regensburg",  # Uncertain due to duplicate number  # Possibly retired
    16: "Pforzheim",
    # 17: "Neumünster",  # Uncertain due to duplicate number
    # 18: "Gelnhausen",  # Probably retired -> 166
    19: "Osnabrück",
    20: "Lüneburg",
    # 53: "Neumünster",  # Uncertain due to duplicate number
    58: "Gütersloh",
    60: "Mühlheim an der Ruhr",
    61: "Bebra",
    63: None,
    64: None,
    65: None,
    68: "Crailsheim",
    # 69: "Worms",  # Probably retired -> 169
    70: None,
    72: "Aschaffenburg",
    73: "Basel",  # Possible duplicate number -> 173
    74: "Zürich",
    75: None,
    76: "Interlaken",
    77: None,
    # 81: "Regensburg",  # Uncertain due to duplicate number  # Possibly retired
    90: None,
    102: "Jever",
    104: "Fulda",
    108: "Lichtenfels",
    # 115: "Regensburg",  # Uncertain due to duplicate number  # Possibly retired
    152: "Hanau",
    154: "Flensburg",
    155: "Rosenheim",
    156: "Heppenheim/Bergstraße",
    157: "Landshut",
    166: "Gelnhausen",
    169: "Worms",
    171: "Heusenstamm",
    # 173: "Basel",  # Probably retired -> 73
    178: "Bremerhaven",
    179: None,
    180: "Castrop Rauxel",
    182: "Rüdesheim am Rhein",
    183: "Timmendorfer Strand",
    185: "Freilassing",
    186: "Chur",
    187: "Mühldorf a. Inn",
    188: "Hildesheim",
    189: None,
    201: "Rheinsberg",
    202: "Wuppertal",
    203: "Cottbus/Chóśebuz",
    204: "Bielefeld",
    # 205: "Zwickau", # number changed to 225
    206: "Magdeburg",
    207: "Stendal",
    208: "Bonn",
    210: "Fontanestadt Neuruppin",
    211: "Uelzen",
    212: "Potsdam",
    213: "Nauen",
    214: "Hamm (Westf.)",
    215: "Bitterfeld-Wolfen",
    216: "Dessau",
    217: "Bergen auf Rügen",
    218: "Braunschweig",
    219: "Hagen",
    220: "Meiningen",
    221: "Lübbenau/Spreewald",
    222: "Eberswalde",
    224: "Saalfeld (Saale)",
    225: "Zwickau",  # previously Oldenburg (Oldb)
    226: "Lutherstadt Wittenberg",
    227: "Ludwigslust",
    228: "Altenburg",
    229: "Templin",
    230: "Delitzsch",
    231: "Brandenburg an der Havel",
    232: "Frankfurt (Oder)",
    233: "Ulm",
    234: "Minden",
    235: "Görlitz",
    236: "Jüterbog",
    237: "Neustrelitz",
    238: "Saarbrücken",
    239: "Essen",
    240: "Bochum",
    241: "Bad Hersfeld ",
    242: "Quedlinburg",
    243: "Bautzen/Budyšin",
    244: "Koblenz",
    302: "Hansestadt Lübeck",  # previously Wiesbaden
    303: "Dortmund",
    304: "München",
    305: "Baden-Baden",
    308: "Murnau am Staffelsee",
    309: "Aalen",
    310: "Wolfsburg",  # possibly "München / Wolfsburg"
    # 311: "Hansestadt Lübeck", # number changed to 302
    312: "Montabaur",
    313: "Treuchtlingen",
    314: "Bergisch Gladbach",
    315: "Singen (Hohentwiel)",
    319: "Duisburg",
    320: "Weil am Rhein",
    323: "Schaffhausen",
    324: "Fürth",
    327: "Siegen",
    328: "Aachen",
    329: None,
    330: "Göttingen",
    331: "Westerland/Sylt",
    332: "Augsburg",
    333: "Goslar",
    334: "Offenburg",
    335: "Konstanz",
    336: "Ingolstadt",
    337: "Stuttgart",
    352: "Mönchengladbach",
    354: "Mittenwald",
    355: "Tuttlingen",
    356: None,
    357: "Esslingen am Neckar",
    358: "St. Ingbert",
    360: "Linz am Rhein",
    361: "Celle",
    363: "Weilheim i.OB",
    701: None,
    702: None,
    703: None,
    704: None,
    705: None,
    707: None,
    708: None,
    710: "Ansbach",
    711: None,
    712: "Dillingen a.d.Donau",
    713: None,
    717: "Paris",
    1101: "Neustadt an der Weinstraße",
    1102: "Neubrandenburg",
    1103: "Paderborn",
    1104: "Erfurt",
    1105: "Dresden",
    1107: "Pirna",
    1108: "Berlin",
    1109: "Güstrow",
    1111: "Hansestadt Wismar",
    1112: "Freie und Hansestadt Hamburg",
    1113: "Hansestadt Stralsund",
    1117: "Erlangen",
    1118: "Plauen/Vogtland",
    1125: "Arnstadt",
    1126: "Leipzig",
    1127: "Weimar",
    1128: "Reutlingen",
    1129: "Kiel",
    1130: "Jena",
    1131: "Trier",
    1132: "Wittenberge",
    1151: "Elsterwerda",
    1152: "Travemünde",
    1153: "Ilmenau",
    1154: "Sonneberg",
    1155: "Mühlhausen/Thüringen",
    1156: "Waren (Müritz)",
    1157: "Innsbruck",
    1158: "Falkenberg/Elster",
    1159: "Passau",
    1160: "Markt Holzkirchen",
    1161: "Andernach",
    1162: "Vaihingen an der Enz",
    1164: "Rödental",
    1165: "Bad Oeynhausen",
    1166: "Bingen am Rhein",
    1167: "Traunstein",
    1168: "Ellwangen",
    1169: "Tutzing",
    1170: "Prenzlau",
    1171: "Oschatz",
    1172: "Bamberg",
    1173: "Halle (Saale)",
    1174: "Hansestadt Warburg",
    1175: "Villingen-Schwenningen",
    1176: "Coburg",
    1177: "Rathenow",
    1178: "Ostseebad Warnemünde",
    1180: "Darmstadt",
    1182: "Mainz",
    1183: "Oberursel (Taunus)",
    1184: "Kaiserslautern",
    1190: "Wien",
    1191: "Salzburg",
    1192: "Linz",
    1501: "Eisenach",
    1502: "Karlsruhe",
    1503: "Altenbeken",
    1504: "Heidelberg",
    1505: "Marburg/Lahn",
    1506: "Kassel",
    1520: "Gotha",
    1521: "Homburg/Saar",
    1522: "Torgau",
    1523: "Hansestadt Greifswald",
    1524: "Hansestadt Rostock",
    4602: "Euregio Maas-Rhein",
    4603: "Mannheim",
    4610: "Frankfurt am Main",
    4685: "Schwäbisch Hall",
    8006: None,
    8008: None,
    8009: None,
    8011: None,
    8012: None,
    8015: None,
    8016: None,
    9001: None,
    9002: None,
    9003: None,
    9004: None,
    9005: None,
    9006: "Martin Luther",
    9007: None,
    9008: None,
    9009: None,
    9010: None,
    9011: None,
    9012: None,
    9013: None,
    9014: None,
    9015: None,
    9016: None,
    9017: None,
    9018: "Freistaat Bayern",
    9019: None,
    9020: None,
    9021: None,
    9022: None,
    9023: None,
    9024: None,
    9025: "Nordrhein-Westfalen",
    9026: "Zürichsee",
    9027: None,
    9028: "Freistaat Sachsen",
    9029: None,
    9030: None,
    9031: None,
    9032: None,
    9033: None,
    9034: None,
    9035: None,
    9036: None,
    9037: None,
    9038: None,
    9039: None,
    9040: None,
    9041: "Baden-Württemberg",
    9042: None,
    9043: None,
    9044: None,
    9045: None,
    9046: None,
    9047: None,
    9048: None,
    9049: None,
    9050: "Metropole Ruhr",
    9201: None,
    9202: "Schleswig-Holstein",
    9204: None,
    9205: None,
    9206: None,
    9209: None,
    9210: None,
    9211: None,
    9212: None,
    9213: None,
    9214: None,
    9215: None,
    9216: None,
    9217: None,
    9218: None,
    9219: None,
    9220: None,
    9221: None,
    9224: None,
    9225: None,
    9226: None,
    9228: None,
    9229: None,
    9230: None,
    9234: None,
    9235: None,
    9237: "Spree",
    9451: None,
    9452: None,
    9453: None,
    9454: None,
    9455: None,
    9456: None,
    9457: "Bundesrepublik Deutschland / Wendlingen-Ulm",
    9458: None,
    9459: None,
    9460: None,
    9461: None,
    9462: None,
    9463: None,
    9464: None,
    9465: None,
    9466: None,
    9467: None,
    9468: None,
    9469: None,
    9470: None,
    9471: None,
    9472: None,
    9473: None,
    9474: None,
    9475: None,
    9477: None,
    9478: None,
    9479: None,
    9480: None,
    9481: "Rheinland-Pfalz",
    9482: None,
    9483: None,
    9484: None,
    9485: None,
    9486: None,
    9487: None,
    9488: None,
    9490: None,
    9491: None,
    9492: None,
    9494: None,
    9495: None,
    9496: None,
    9497: None,
    9498: None,
    9500: None,
}
"""
A dict containing the train ID and its corresponding name that is printed on the side of the trains first and last coach
"""



"""
Package for Austrian trains.

Contains the following operators:

- ``onboardapis.train.at.obb`` - Österreichische Bundesbahnen

---
"""
from __future__ import annotations

__all__ = [
    "obb",
]
